@@ -8,17 +8,21 @@ use std::path::{Path, PathBuf};
 use std::process::{Command, Stdio};
 use std::ptr;
 
-use anyhow::{bail, Context, Result};
+use anyhow::{bail, ensure, Context, Result};
 use memmap::Mmap;
 
 use crate::btf;
 use crate::metadata;
 use crate::metadata::UnprocessedObj;
 
-enum OutputDest<'a> {
+pub enum OutputDest<'a> {
     Stdout,
     /// Infer a filename and place file in specified directory
     Directory(&'a Path),
+    #[allow(dead_code)]
+    /// File to place output in
+    // Only constructed in libbpf-cargo library
+    File(&'a Path),
 }
 
 macro_rules! gen_bpf_object_iter {
@@ -71,10 +75,19 @@ fn rustfmt(s: &str, rustfmt_path: Option<&PathBuf>) -> Result<String> {
     .stdout(Stdio::piped())
     .spawn()
     .context("Failed to spawn rustfmt")?;
+
+    // Send input in via stdin
     write!(cmd.stdin.take().unwrap(), "{}", s)?;
+
+    // Extract output
     let output = cmd
         .wait_with_output()
         .context("Failed to execute rustfmt")?;
+    ensure!(
+        output.status.success(),
+        "Failed to rustfmt: {}",
+        output.status
+    );
 
     Ok(String::from_utf8(output.stdout)?)
 }
@@ -139,12 +152,15 @@ fn get_prog_name(prog: *const libbpf_sys::bpf_program) -> Result<String> {
 }
 
 fn map_is_mmapable(map: *const libbpf_sys::bpf_map) -> bool {
-    let internal = unsafe { libbpf_sys::bpf_map__is_internal(map) };
     let def = unsafe { libbpf_sys::bpf_map__def(map) };
-    let mmapable = unsafe { (*def).map_flags } & libbpf_sys::BPF_F_MMAPABLE;
-    let name = get_map_name(map);
+    (unsafe { (*def).map_flags } & libbpf_sys::BPF_F_MMAPABLE) > 0
+}
 
-    internal && (mmapable > 0) && (name.is_ok() && name.unwrap().is_some())
+fn map_is_datasec(map: *const libbpf_sys::bpf_map) -> bool {
+    let internal = unsafe { libbpf_sys::bpf_map__is_internal(map) };
+    let mmapable = map_is_mmapable(map);
+
+    internal && mmapable
 }
 
 fn map_is_readonly(map: *const libbpf_sys::bpf_map) -> bool {
@@ -220,25 +236,27 @@ fn gen_skel_map_defs(
     object: *mut libbpf_sys::bpf_object,
     obj_name: &str,
     open: bool,
+    mutable: bool,
 ) -> Result<()> {
-    // If no non-datasec maps, return early
-    if MapIter::new(object)
-        .filter(|map| !map_is_mmapable(*map))
-        .count()
-        == 0
-    {
+    if MapIter::new(object).next().is_none() {
         return Ok(());
     }
 
+    let (struct_suffix, mut_prefix, map_fn) = if mutable {
+        ("Mut", "mut ", "map_mut")
+    } else {
+        ("", "", "map")
+    };
+
     let (struct_name, inner_ty, return_ty) = if open {
         (
-            format!("Open{}Maps", obj_name),
+            format!("Open{}Maps{}", obj_name, struct_suffix),
             "libbpf_rs::OpenObject",
             "libbpf_rs::OpenMap",
         )
     } else {
         (
-            format!("{}Maps", obj_name),
+            format!("{}Maps{}", obj_name, struct_suffix),
             "libbpf_rs::Object",
             "libbpf_rs::Map",
         )
@@ -248,20 +266,17 @@ fn gen_skel_map_defs(
         skel,
         r#"
         pub struct {struct_name}<'a> {{
-            inner: &'a mut {inner_ty},
+            inner: &'a {mut_prefix}{inner_ty},
         }}
 
         impl<'a> {struct_name}<'a> {{
         "#,
         inner_ty = inner_ty,
-        struct_name = struct_name
+        struct_name = struct_name,
+        mut_prefix = mut_prefix,
     )?;
 
     for map in MapIter::new(object) {
-        if map_is_mmapable(map) {
-            continue;
-        }
-
         let map_name = match get_map_name(map)? {
             Some(n) => n,
             None => continue,
@@ -270,13 +285,15 @@ fn gen_skel_map_defs(
         write!(
             skel,
             r#"
-            pub fn {map_name}(&mut self) -> &mut {return_ty} {{
-                self.inner.map_unwrap("{raw_map_name}")
+            pub fn {map_name}(&{mut_prefix}self) -> &{mut_prefix}{return_ty} {{
+                self.inner.{map_fn}("{raw_map_name}").unwrap()
             }}
             "#,
             map_name = map_name,
             raw_map_name = get_raw_map_name(map)?,
             return_ty = return_ty,
+            mut_prefix = mut_prefix,
+            map_fn = map_fn
         )?;
     }
 
@@ -290,20 +307,27 @@ fn gen_skel_prog_defs(
     object: *mut libbpf_sys::bpf_object,
     obj_name: &str,
     open: bool,
+    mutable: bool,
 ) -> Result<()> {
     if ProgIter::new(object).next().is_none() {
         return Ok(());
     }
 
+    let (struct_suffix, mut_prefix, prog_fn) = if mutable {
+        ("Mut", "mut ", "prog_mut")
+    } else {
+        ("", "", "prog")
+    };
+
     let (struct_name, inner_ty, return_ty) = if open {
         (
-            format!("Open{}Progs", obj_name),
+            format!("Open{}Progs{}", obj_name, struct_suffix),
             "libbpf_rs::OpenObject",
             "libbpf_rs::OpenProgram",
         )
     } else {
         (
-            format!("{}Progs", obj_name),
+            format!("{}Progs{}", obj_name, struct_suffix),
             "libbpf_rs::Object",
             "libbpf_rs::Program",
         )
@@ -313,25 +337,28 @@ fn gen_skel_prog_defs(
         skel,
         r#"
         pub struct {struct_name}<'a> {{
-            inner: &'a mut {inner_ty},
+            inner: &'a {mut_prefix}{inner_ty},
         }}
 
         impl<'a> {struct_name}<'a> {{
         "#,
         inner_ty = inner_ty,
-        struct_name = struct_name
+        struct_name = struct_name,
+        mut_prefix = mut_prefix,
     )?;
 
     for prog in ProgIter::new(object) {
         write!(
             skel,
             r#"
-            pub fn {prog_name}(&mut self) -> &mut {return_ty} {{
-                self.inner.prog_unwrap("{prog_name}")
+            pub fn {prog_name}(&{mut_prefix}self) -> &{mut_prefix}{return_ty} {{
+                self.inner.{prog_fn}("{prog_name}").unwrap()
             }}
             "#,
             prog_name = get_prog_name(prog)?,
             return_ty = return_ty,
+            mut_prefix = mut_prefix,
+            prog_fn = prog_fn
         )?;
     }
 
@@ -376,32 +403,36 @@ fn gen_skel_map_getter(
     object: *mut libbpf_sys::bpf_object,
     obj_name: &str,
     open: bool,
+    mutable: bool,
 ) -> Result<()> {
-    // If no non-datasec maps, return early
-    if MapIter::new(object)
-        .filter(|map| !map_is_mmapable(*map))
-        .count()
-        == 0
-    {
+    if MapIter::new(object).next().is_none() {
         return Ok(());
     }
 
-    let return_ty = if open {
-        format!("Open{}Maps", obj_name)
+    let (struct_suffix, mut_prefix, map_fn) = if mutable {
+        ("Mut", "mut ", "maps_mut")
     } else {
-        format!("{}Maps", obj_name)
+        ("", "", "maps")
+    };
+
+    let return_ty = if open {
+        format!("Open{}Maps{}", obj_name, struct_suffix)
+    } else {
+        format!("{}Maps{}", obj_name, struct_suffix)
     };
 
     write!(
         skel,
         r#"
-        pub fn maps(&mut self) -> {return_ty} {{
+        pub fn {map_fn}(&{mut_prefix}self) -> {return_ty} {{
             {return_ty} {{
-                inner: &mut self.obj,
+                inner: &{mut_prefix}self.obj,
             }}
         }}
         "#,
-        return_ty = return_ty
+        return_ty = return_ty,
+        map_fn = map_fn,
+        mut_prefix = mut_prefix,
     )?;
 
     Ok(())
@@ -412,27 +443,36 @@ fn gen_skel_prog_getter(
     object: *mut libbpf_sys::bpf_object,
     obj_name: &str,
     open: bool,
+    mutable: bool,
 ) -> Result<()> {
     if ProgIter::new(object).next().is_none() {
         return Ok(());
     }
 
-    let return_ty = if open {
-        format!("Open{}Progs", obj_name)
+    let (struct_suffix, mut_prefix, prog_fn) = if mutable {
+        ("Mut", "mut ", "progs_mut")
     } else {
-        format!("{}Progs", obj_name)
+        ("", "", "progs")
+    };
+
+    let return_ty = if open {
+        format!("Open{}Progs{}", obj_name, struct_suffix)
+    } else {
+        format!("{}Progs{}", obj_name, struct_suffix)
     };
 
     write!(
         skel,
         r#"
-        pub fn progs(&mut self) -> {return_ty} {{
+        pub fn {prog_fn}(&{mut_prefix}self) -> {return_ty} {{
             {return_ty} {{
-                inner: &mut self.obj,
+                inner: &{mut_prefix}self.obj,
             }}
         }}
         "#,
-        return_ty = return_ty
+        return_ty = return_ty,
+        mut_prefix = mut_prefix,
+        prog_fn = prog_fn,
     )?;
 
     Ok(())
@@ -445,7 +485,7 @@ fn gen_skel_datasec_getters(
     loaded: bool,
 ) -> Result<()> {
     for (idx, map) in MapIter::new(object).enumerate() {
-        if !map_is_mmapable(map) {
+        if !map_is_datasec(map) {
             continue;
         }
 
@@ -662,7 +702,7 @@ fn gen_skel_contents(_debug: bool, raw_obj_name: &str, obj_file_path: &Path) ->
                     return Err(libbpf_rs::Error::System(-ret));
                 }}
 
-                let obj = unsafe {{ libbpf_rs::OpenObject::from_ptr(skel_config.object_ptr()) }};
+                let obj = unsafe {{ libbpf_rs::OpenObject::from_ptr(skel_config.object_ptr())? }};
 
                 Ok(Open{name}Skel {{
                     obj,
@@ -674,8 +714,10 @@ fn gen_skel_contents(_debug: bool, raw_obj_name: &str, obj_file_path: &Path) ->
         name = obj_name
     )?;
 
-    gen_skel_map_defs(&mut skel, object, &obj_name, true)?;
-    gen_skel_prog_defs(&mut skel, object, &obj_name, true)?;
+    gen_skel_map_defs(&mut skel, object, &obj_name, true, false)?;
+    gen_skel_map_defs(&mut skel, object, &obj_name, true, true)?;
+    gen_skel_prog_defs(&mut skel, object, &obj_name, true, false)?;
+    gen_skel_prog_defs(&mut skel, object, &obj_name, true, true)?;
     gen_skel_datasec_defs(&mut skel, raw_obj_name, &*mmap)?;
 
     write!(
@@ -693,7 +735,7 @@ fn gen_skel_contents(_debug: bool, raw_obj_name: &str, obj_file_path: &Path) ->
                     return Err(libbpf_rs::Error::System(-ret));
                 }}
 
-                let obj = unsafe {{ libbpf_rs::Object::from_ptr(self.obj.take_ptr()) }};
+                let obj = unsafe {{ libbpf_rs::Object::from_ptr(self.obj.take_ptr())? }};
 
                 Ok({name}Skel {{
                     obj,
@@ -709,13 +751,17 @@ fn gen_skel_contents(_debug: bool, raw_obj_name: &str, obj_file_path: &Path) ->
             "".to_string()
         }
     )?;
-    gen_skel_prog_getter(&mut skel, object, &obj_name, true)?;
-    gen_skel_map_getter(&mut skel, object, &obj_name, true)?;
+    gen_skel_prog_getter(&mut skel, object, &obj_name, true, false)?;
+    gen_skel_prog_getter(&mut skel, object, &obj_name, true, true)?;
+    gen_skel_map_getter(&mut skel, object, &obj_name, true, false)?;
+    gen_skel_map_getter(&mut skel, object, &obj_name, true, true)?;
     gen_skel_datasec_getters(&mut skel, object, raw_obj_name, false)?;
     writeln!(skel, "}}")?;
 
-    gen_skel_map_defs(&mut skel, object, &obj_name, false)?;
-    gen_skel_prog_defs(&mut skel, object, &obj_name, false)?;
+    gen_skel_map_defs(&mut skel, object, &obj_name, false, false)?;
+    gen_skel_map_defs(&mut skel, object, &obj_name, false, true)?;
+    gen_skel_prog_defs(&mut skel, object, &obj_name, false, false)?;
+    gen_skel_prog_defs(&mut skel, object, &obj_name, false, true)?;
     gen_skel_link_defs(&mut skel, object, &obj_name)?;
 
     write!(
@@ -737,8 +783,10 @@ fn gen_skel_contents(_debug: bool, raw_obj_name: &str, obj_file_path: &Path) ->
         "#,
         name = &obj_name,
     )?;
-    gen_skel_prog_getter(&mut skel, object, &obj_name, false)?;
-    gen_skel_map_getter(&mut skel, object, &obj_name, false)?;
+    gen_skel_prog_getter(&mut skel, object, &obj_name, false, false)?;
+    gen_skel_prog_getter(&mut skel, object, &obj_name, false, true)?;
+    gen_skel_map_getter(&mut skel, object, &obj_name, false, false)?;
+    gen_skel_map_getter(&mut skel, object, &obj_name, false, true)?;
     gen_skel_datasec_getters(&mut skel, object, raw_obj_name, true)?;
     gen_skel_attach(&mut skel, object, &obj_name)?;
     writeln!(skel, "}}")?;
@@ -777,6 +825,10 @@ fn gen_skel(
             let mut file = File::create(path)?;
             file.write_all(skel.as_bytes())?;
         }
+        OutputDest::File(file) => {
+            let mut file = File::create(file)?;
+            file.write_all(skel.as_bytes())?;
+        }
     };
 
     Ok(())
@@ -832,71 +884,58 @@ pub fn gen_mods(objs: &[UnprocessedObj], rustfmt_path: Option<&PathBuf>) -> Resu
     Ok(())
 }
 
-fn gen_single(debug: bool, obj_file: &Path, rustfmt_path: Option<&PathBuf>) -> i32 {
+pub fn gen_single(
+    debug: bool,
+    obj_file: &Path,
+    output: OutputDest,
+    rustfmt_path: Option<&PathBuf>,
+) -> Result<()> {
     let filename = match obj_file.file_name() {
         Some(n) => n,
-        None => {
-            eprintln!(
-                "Could not determine file name for object file: {}",
-                obj_file.to_string_lossy()
-            );
-            return 1;
-        }
+        None => bail!(
+            "Could not determine file name for object file: {}",
+            obj_file.to_string_lossy()
+        ),
     };
 
     let name = match filename.to_str() {
         Some(n) => {
             if !n.ends_with(".o") {
-                eprintln!("Object file does not have `.o` suffix: {}", n);
-                return 1;
+                bail!("Object file does not have `.o` suffix: {}", n);
             }
 
             n.split('.').next().unwrap()
         }
-        None => {
-            eprintln!(
-                "Object file name is not valid unicode: {}",
-                filename.to_string_lossy()
-            );
-            return 1;
-        }
+        None => bail!(
+            "Object file name is not valid unicode: {}",
+            filename.to_string_lossy()
+        ),
     };
 
-    match gen_skel(debug, name, obj_file, OutputDest::Stdout, rustfmt_path) {
-        Ok(_) => 0,
-        Err(e) => {
-            eprintln!(
-                "Failed to generate skeleton for {}: {}",
-                obj_file.to_string_lossy(),
-                e
-            );
-
-            1
-        }
+    if let Err(e) = gen_skel(debug, name, obj_file, output, rustfmt_path) {
+        bail!(
+            "Failed to generate skeleton for {}: {}",
+            obj_file.to_string_lossy(),
+            e
+        );
     }
+
+    Ok(())
 }
 
 fn gen_project(
     debug: bool,
     manifest_path: Option<&PathBuf>,
     rustfmt_path: Option<&PathBuf>,
-) -> i32 {
-    let to_gen = match metadata::get(debug, manifest_path) {
-        Ok(v) => v,
-        Err(e) => {
-            eprintln!("{}", e);
-            return 1;
-        }
-    };
-
+) -> Result<()> {
+    let to_gen = metadata::get(debug, manifest_path)?;
     if debug && !to_gen.is_empty() {
         println!("Found bpf objs to gen skel:");
         for obj in &to_gen {
             println!("\t{:?}", obj);
         }
     } else if to_gen.is_empty() {
-        eprintln!("Did not find any bpf objects to generate skeleton");
-        return 1;
+        bail!("Did not find any bpf objects to generate skeleton");
     }
 
     // Map to store package_name -> [UnprocessedObj]
@@ -917,14 +956,11 @@ fn gen_project(
             rustfmt_path,
         ) {
             Ok(_) => (),
-            Err(e) => {
-                eprintln!(
-                    "Failed to generate skeleton for {}: {}",
-                    obj.path.as_path().display(),
-                    e
-                );
-                return 1;
-            }
+            Err(e) => bail!(
+                "Failed to generate skeleton for {}: {}",
+                obj.path.as_path().display(),
+                e
+            ),
         }
 
         match package_objs.get_mut(&obj.package) {
@@ -936,16 +972,12 @@ fn gen_project(
     }
 
     for (package, objs) in package_objs {
-        match gen_mods(&objs, rustfmt_path) {
-            Ok(_) => (),
-            Err(e) => {
-                eprintln!("Failed to generate mod.rs for package={}: {}", package, e);
-                return 1;
-            }
+        if let Err(e) = gen_mods(&objs, rustfmt_path) {
+            bail!("Failed to generate mod.rs for package={}: {}", package, e);
         }
     }
 
-    0
+    Ok(())
 }
 
 pub fn gen(
@@ -953,14 +985,13 @@ pub fn gen(
     manifest_path: Option<&PathBuf>,
     rustfmt_path: Option<&PathBuf>,
     object: Option<&PathBuf>,
-) -> i32 {
+) -> Result<()> {
     if manifest_path.is_some() && object.is_some() {
-        eprintln!("--manifest-path and --object cannot be used together");
-        return 1;
+        bail!("--manifest-path and --object cannot be used together");
     }
 
     if let Some(obj_file) = object {
-        gen_single(debug, obj_file, rustfmt_path)
+        gen_single(debug, obj_file, OutputDest::Stdout, rustfmt_path)
     } else {
         gen_project(debug, manifest_path, rustfmt_path)
     }
